@@ -27,37 +27,46 @@ from genshi.builder import tag
 from trac.config import Configuration, ConfigSection
 from trac.core import *
 from trac.env import IEnvironmentSetupParticipant
-from trac.perm import PermissionSystem
+from trac.perm import PermissionCache, PermissionSystem
+from trac.resource import ResourceNotFound
 from trac.ticket.api import ITicketActionController, TicketSystem
-from trac.ticket.model import Resolution
-from trac.util import get_reporter_id, to_list
+from trac.ticket.model import Component as TicketComponent, Resolution
+from trac.util import get_reporter_id, sub_val, to_list
 from trac.util.presentation import separated
 from trac.util.translation import _, tag_, cleandoc_
 from trac.web.chrome import Chrome, add_script, add_script_data
 from trac.wiki.formatter import system_message
 from trac.wiki.macros import WikiMacroBase
 
+
 # -- Utilities for the ConfigurableTicketWorkflow
 
 def parse_workflow_config(rawactions):
     """Given a list of options from [ticket-workflow]"""
 
-    default = {
+    required_attrs = {
         'oldstates': [],
         'newstate': '',
         'name': '',
+        'label': '',
         'default': 0,
         'operations': [],
-        'permissions': []
+        'permissions': [],
     }
+    optional_attrs = {
+        'set_owner': [],
+        'set_resolution': [],
+    }
+    known_attrs = required_attrs.copy()
+    known_attrs.update(optional_attrs)
 
-    actions = defaultdict(lambda: default.copy())
+    actions = defaultdict(dict)
     for option, value in rawactions:
         parts = option.split('.')
         name = parts[0]
         if len(parts) == 1:
-            # Base name, of the syntax: old,states,here -> newstate
             try:
+                # Base name, of the syntax: old,states,here -> newstate
                 oldstates, newstate = [x.strip() for x in value.split('->')]
             except ValueError:
                 continue  # Syntax error, a warning will be logged later
@@ -65,15 +74,23 @@ def parse_workflow_config(rawactions):
             actions[name]['newstate'] = newstate
         else:
             attribute = parts[1]
-            if attribute == 'default':
-                actions[name][attribute] = int(value)
-            elif attribute in ('operations', 'permissions'):
-                actions[name][attribute] = to_list(value)
-            else:
+            if attribute not in known_attrs.keys() or \
+                    isinstance(known_attrs[attribute], str):
                 actions[name][attribute] = value
-    for name, attrs in actions.iteritems():
-        if not attrs.get('name'):
-            attrs['name'] = name
+            elif isinstance(known_attrs[attribute], int):
+                actions[name][attribute] = int(value)
+            elif isinstance(known_attrs[attribute], list):
+                actions[name][attribute] = to_list(value)
+
+    for action, attributes in actions.items():
+        if 'label' not in attributes:
+            if 'name' in attributes:  # backwards-compatibility, #11828
+                attributes['label'] = attributes['name']
+            else:
+                attributes['label'] = action.replace("_", " ").strip()
+        for key, val in required_attrs.items():
+            attributes.setdefault(key, val)
+
     return actions
 
 
@@ -85,6 +102,7 @@ def get_workflow_config(config):
     actions = parse_workflow_config(raw_actions)
     return actions
 
+
 def load_workflow_config_snippet(config, filename):
     """Loads the ticket-workflow section from the given file (expected to be in
     the 'workflows' tree) into the provided config.
@@ -110,10 +128,9 @@ class ConfigurableTicketWorkflow(Component):
         there's only a `ConfigurableTicketWorkflow` component in charge.
         That component allows the workflow to be configured via this section
         in the `trac.ini` file. See TracWorkflow for more details.
+        """)
 
-        (''since 0.11'')""")
-
-    def __init__(self, *args, **kwargs):
+    def __init__(self):
         self.actions = self.get_all_actions()
         self.log.debug('Workflow actions at initialization: %s\n',
                        self.actions)
@@ -129,13 +146,13 @@ class ConfigurableTicketWorkflow(Component):
             self.config.save()
             self.actions = self.get_all_actions()
 
-    def environment_needs_upgrade(self, db):
+    def environment_needs_upgrade(self):
         """The environment needs an upgrade if there is no [ticket-workflow]
         section in the config.
         """
         return not list(self.config.options('ticket-workflow'))
 
-    def upgrade_environment(self, db):
+    def upgrade_environment(self):
         """Insert a [ticket-workflow] section using the original-workflow"""
         load_workflow_config_snippet(self.config, 'original-workflow.ini')
         self.config.save()
@@ -154,7 +171,7 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
 
 """
         self.log.info(info_message.replace('\n', ' ').replace('==', ''))
-        print info_message
+        print(info_message)
 
     # ITicketActionController methods
 
@@ -167,25 +184,25 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
         # the process of being modified, we need to base our information on the
         # pre-modified state so that we don't try to do two (or more!) steps at
         # once and get really confused.
-        status = ticket._old.get('status', ticket['status']) or 'new'
+        status = ticket._old.get('status', ticket['status'])
+        exists = status is not None
 
         ticket_perm = req.perm(ticket.resource)
         allowed_actions = []
         for action_name, action_info in self.actions.items():
             oldstates = action_info['oldstates']
-            if oldstates == ['*'] or status in oldstates:
+            if exists and oldstates == ['*'] or status in oldstates:
                 # This action is valid in this state.  Check permissions.
                 required_perms = action_info['permissions']
                 if self._is_action_allowed(ticket_perm, required_perms):
                     allowed_actions.append((action_info['default'],
                                             action_name))
         # Append special `_reset` action if status is invalid.
-        if status not in TicketSystem(self.env).get_all_status() + \
-                         ['new', 'closed']:
-            required_perms = self.actions['_reset'].get('permissions')
+        if exists and status not in TicketSystem(self.env).get_all_status():
+            reset = self.actions['_reset']
+            required_perms = reset['permissions']
             if self._is_action_allowed(ticket_perm, required_perms):
-                default = self.actions['_reset'].get('default')
-                allowed_actions.append((default, '_reset'))
+                allowed_actions.append((reset['default'], '_reset'))
         return allowed_actions
 
     def _is_action_allowed(self, ticket_perm, required_perms):
@@ -206,6 +223,7 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
             all_status.add(attributes['newstate'])
         all_status.discard('*')
         all_status.discard('')
+        all_status.discard(None)
         return all_status
 
     def render_ticket_action_control(self, req, ticket, action):
@@ -217,8 +235,12 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
         operations = this_action['operations']
         current_owner = ticket._old.get('owner', ticket['owner'])
         author = get_reporter_id(req, 'author')
-        format_author = partial(Chrome(self.env).format_author, req)
-        formatted_current_owner = format_author(current_owner or _("(none)"))
+        author_info = partial(Chrome(self.env).authorinfo, req,
+                              resource=ticket.resource)
+        format_author = partial(Chrome(self.env).format_author, req,
+                                resource=ticket.resource)
+        formatted_current_owner = author_info(current_owner)
+        exists = ticket._old.get('status', ticket['status']) is not None
 
         control = []  # default to nothing
         hints = []
@@ -227,58 +249,92 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
             hints.append(_("Current state no longer exists"))
         if 'del_owner' in operations:
             hints.append(_("The ticket will be disowned"))
-        if 'set_owner' in operations:
-            id = 'action_%s_reassign_owner' % action
-
+        if 'set_owner' in operations or 'may_set_owner' in operations:
             if 'set_owner' in this_action:
-                owners = [x.strip() for x in
-                          this_action['set_owner'].split(',')]
+                owners = self._to_users(this_action['set_owner'], ticket)
             elif self.config.getbool('ticket', 'restrict_owner'):
                 perm = PermissionSystem(self.env)
                 owners = perm.get_users_with_permission('TICKET_MODIFY')
-                owners.sort()
+                owners = [user for user in owners
+                               if 'TICKET_MODIFY'
+                               in PermissionCache(self.env, user,
+                                                  ticket.resource)]
+                owners = sorted(owners)
             else:
                 owners = None
 
-            if owners is None:
-                owner = req.args.get(id, author)
-                control.append(tag_("to %(owner)s",
-                                    owner=tag.input(type='text', id=id,
-                                                    name=id, value=owner)))
-                hints.append(_("The owner will be changed from "
-                               "%(current_owner)s to the specified user",
-                               current_owner=formatted_current_owner))
+            if 'set_owner' in operations:
+                default_owner = author
+            elif 'may_set_owner' in operations:
+                if not exists:
+                    default_owner = TicketSystem(self.env).default_owner
+                else:
+                    default_owner = ticket._old.get('owner',
+                                                    ticket['owner'] or None)
+                if owners is not None and default_owner not in owners:
+                    owners.insert(0, default_owner)
+            else:
+                # Protect against future modification for case that another
+                # operation is added to the outer conditional
+                raise AssertionError(operations)
+
+            id = 'action_%s_reassign_owner' % action
+
+            if not owners:
+                owner = req.args.get(id, default_owner)
+                control.append(
+                    tag_("to %(owner)s",
+                         owner=tag.input(type='text', id=id, name=id,
+                                         value=owner)))
+                if not exists or current_owner is None:
+                    hints.append(_("The owner will be the specified user"))
+                else:
+                    hints.append(tag_("The owner will be changed from "
+                                      "%(current_owner)s to the specified "
+                                      "user",
+                                      current_owner=formatted_current_owner))
             elif len(owners) == 1:
                 owner = tag.input(type='hidden', id=id, name=id,
                                   value=owners[0])
-                formatted_new_owner = format_author(owners[0])
+                formatted_new_owner = author_info(owners[0])
                 control.append(tag_("to %(owner)s",
                                     owner=tag(formatted_new_owner, owner)))
-                if ticket['owner'] != owners[0]:
-                    hints.append(_("The owner will be changed from "
-                                   "%(current_owner)s to %(selected_owner)s",
-                                   current_owner=formatted_current_owner,
-                                   selected_owner=formatted_new_owner))
+                if not exists or current_owner is None:
+                    hints.append(tag_("The owner will be %(new_owner)s",
+                                      new_owner=formatted_new_owner))
+                elif ticket['owner'] != owners[0]:
+                    hints.append(tag_("The owner will be changed from "
+                                      "%(current_owner)s to %(new_owner)s",
+                                      current_owner=formatted_current_owner,
+                                      new_owner=formatted_new_owner))
             else:
-                selected_owner = req.args.get(id, req.authname)
+                selected_owner = req.args.get(id, default_owner)
                 control.append(tag_("to %(owner)s", owner=tag.select(
-                    [tag.option(x, value=x,
-                                selected=(x == selected_owner or None))
-                     for x in owners],
+                    [tag.option(label, value=value if value is not None else '',
+                                selected=(value == selected_owner or None))
+                     for label, value in sorted((format_author(owner), owner)
+                                                for owner in owners)],
                     id=id, name=id)))
-                hints.append(_("The owner will be changed from "
-                               "%(current_owner)s to the selected user",
-                               current_owner=formatted_current_owner))
+                if not exists or current_owner is None:
+                    hints.append(_("The owner will be the selected user"))
+                else:
+                    hints.append(tag_("The owner will be changed from "
+                                      "%(current_owner)s to the selected user",
+                                      current_owner=formatted_current_owner))
         elif 'set_owner_to_self' in operations and \
                 ticket._old.get('owner', ticket['owner']) != author:
-            hints.append(_("The owner will be changed from %(current_owner)s "
-                           "to %(authname)s",
-                           current_owner=formatted_current_owner,
-                           authname=format_author(author)))
+            formatted_author = author_info(author)
+            if not exists or current_owner is None:
+                hints.append(tag_("The owner will be %(new_owner)s",
+                                  new_owner=formatted_author))
+            else:
+                hints.append(tag_("The owner will be changed from "
+                                  "%(current_owner)s to %(new_owner)s",
+                                  current_owner=formatted_current_owner,
+                                  new_owner=formatted_author))
         if 'set_resolution' in operations:
             if 'set_resolution' in this_action:
-                resolutions = [x.strip() for x in
-                               this_action['set_resolution'].split(',')]
+                resolutions = this_action['set_resolution']
             else:
                 resolutions = [r.name for r in Resolution.select(self.env)]
             if not resolutions:
@@ -292,8 +348,8 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
                 control.append(tag_("as %(resolution)s",
                                     resolution=tag(resolutions[0],
                                                    resolution)))
-                hints.append(_("The resolution will be set to %(name)s",
-                               name=resolutions[0]))
+                hints.append(tag_("The resolution will be set to %(name)s",
+                                  name=resolutions[0]))
             else:
                 selected_option = req.args.get(id,
                         TicketSystem(self.env).default_resolution)
@@ -307,19 +363,23 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
         if 'del_resolution' in operations:
             hints.append(_("The resolution will be deleted"))
         if 'leave_status' in operations:
-            control.append(_("as %(status)s",
-                             status= ticket._old.get('status',
-                                                     ticket['status'])))
+            control.append(tag_("as %(status)s",
+                                status=ticket._old.get('status',
+                                                       ticket['status'])))
             if len(operations) == 1:
-                hints.append(_("The owner will remain %(current_owner)s",
-                               current_owner=formatted_current_owner)
+                hints.append(tag_("The owner will remain %(current_owner)s",
+                                  current_owner=formatted_current_owner)
                              if current_owner else
                              _("The ticket will remain with no owner"))
         else:
-            if status != '*':
-                hints.append(_("Next status will be '%(name)s'", name=status))
-        return (this_action.get('name', action), tag(separated(control, ' ')),
-                '. '.join(hints) + '.' if hints else '')
+            if ticket['status'] is None:
+                hints.append(tag_("The status will be '%(name)s'",
+                                  name=status))
+            elif status != '*':
+                hints.append(tag_("Next status will be '%(name)s'",
+                                  name=status))
+        return (this_action['label'], tag(separated(control, ' ')),
+                tag(separated(hints, '. ', '.') if hints else ''))
 
     def get_ticket_changes(self, req, ticket, action):
         this_action = self.actions[action]
@@ -339,26 +399,56 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
         for operation in this_action['operations']:
             if operation == 'del_owner':
                 updated['owner'] = ''
-            elif operation == 'set_owner':
+            elif operation in ('set_owner', 'may_set_owner'):
+                set_owner = this_action.get('set_owner')
                 newowner = req.args.get('action_%s_reassign_owner' % action,
-                                    this_action.get('set_owner', '').strip())
+                                        set_owner[0] if set_owner else '')
                 # If there was already an owner, we get a list, [new, old],
                 # but if there wasn't we just get new.
                 if type(newowner) == list:
                     newowner = newowner[0]
-                updated['owner'] = newowner
+                updated['owner'] = self._sub_owner_keyword(newowner, ticket)
             elif operation == 'set_owner_to_self':
                 updated['owner'] = get_reporter_id(req, 'author')
             elif operation == 'del_resolution':
                 updated['resolution'] = ''
             elif operation == 'set_resolution':
-                newresolution = req.args.get('action_%s_resolve_resolution' % \
-                                             action,
-                                this_action.get('set_resolution', '').strip())
+                set_resolution = this_action.get('set_resolution')
+                newresolution = req.args.get('action_%s_resolve_resolution'
+                                             % action,
+                                             set_resolution[0]
+                                             if set_resolution else '')
                 updated['resolution'] = newresolution
 
             # reset_workflow is just a no-op here, so we don't look for it.
             # leave_status is just a no-op here, so we don't look for it.
+
+        # Set owner to component owner for 'new' ticket if:
+        #  - ticket doesn't exist and owner is < default >
+        #  - component is changed
+        #  - owner isn't explicitly changed
+        #  - ticket owner is equal to owner of previous component
+        #  - new component has an owner
+        if not ticket.exists and 'owner' not in updated:
+            updated['owner'] = self._sub_owner_keyword(ticket['owner'], ticket)
+        elif ticket['status'] == 'new' and \
+                'component' in ticket.values and \
+                'component' in ticket._old and \
+                'owner' not in updated:
+            try:
+                old_comp = TicketComponent(self.env, ticket._old['component'])
+            except ResourceNotFound:
+                # If the old component has been removed from the database
+                # we just leave the owner as is.
+                pass
+            else:
+                old_owner = old_comp.owner or ''
+                current_owner = ticket['owner'] or ''
+                if old_owner == current_owner:
+                    new_comp = TicketComponent(self.env, ticket['component'])
+                    if new_comp.owner:
+                        updated['owner'] = new_comp.owner
+
         return updated
 
     def apply_action_side_effects(self, req, ticket, action):
@@ -383,19 +473,20 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
 
         # Special action that gets enabled if the current status no longer
         # exists, as no other action can then change its state. (#5307/#11850)
-        if '_reset' not in actions:
-            reset = {
-                'default': 0,
-                'name': 'reset',
-                'newstate': 'new',
-                'oldstates': [],
-                'operations': ['reset_workflow'],
-                'permissions': ['TICKET_ADMIN']
-            }
-            for key, val in reset.items():
-                actions['_reset'][key] = val
+        reset = {
+            'default': 0,
+            'label': 'reset',
+            'newstate': 'new',
+            'oldstates': [],
+            'operations': ['reset_workflow'],
+            'permissions': ['TICKET_ADMIN']
+        }
+        for key, val in reset.items():
+            actions['_reset'].setdefault(key, val)
 
         for name, info in actions.iteritems():
+            for val in ('<none>', '< none >'):
+                sub_val(actions[name]['oldstates'], val, None)
             if not info['newstate']:
                 self.log.warning("Ticket workflow action '%s' doesn't define "
                                  "any transitions", name)
@@ -427,6 +518,52 @@ Read TracWorkflow for more information (don't forget to 'wiki upgrade' as well)
                       self._has_perms_for_action(req, info, ticket.resource)]
         return actions
 
+    # Internal methods
+
+    def _sub_owner_keyword(self, owner, ticket):
+        """Substitute keywords from the default_owner field.
+
+        < default > -> component owner
+        """
+        if owner in ('< default >', '<default>'):
+            default_owner = ''
+            if ticket['component']:
+                try:
+                    component = TicketComponent(self.env, ticket['component'])
+                except ResourceNotFound:
+                    pass  # No such component exists
+                else:
+                    default_owner = component.owner  # May be empty
+            return default_owner
+        return owner
+
+    def _to_users(self, users_perms_and_groups, ticket):
+        """Finds all users contained in the list of `users_perms_and_groups`
+        by recursive lookup of users when a `group` is encountered.
+        """
+        ps = PermissionSystem(self.env)
+        groups = ps.get_groups_dict()
+
+        def append_owners(users_perms_and_groups):
+            for user_perm_or_group in users_perms_and_groups:
+                if user_perm_or_group == 'authenticated':
+                    owners.update(set(u[0] for u in self.env.get_known_users()))
+                elif user_perm_or_group.isupper():
+                    perm = user_perm_or_group
+                    for user in ps.get_users_with_permission(perm):
+                        if perm in PermissionCache(self.env, user,
+                                                   ticket.resource):
+                            owners.add(user)
+                elif user_perm_or_group not in groups:
+                    owners.add(user_perm_or_group)
+                else:
+                    append_owners(groups[user_perm_or_group])
+
+        owners = set()
+        append_owners(users_perms_and_groups)
+
+        return sorted(owners)
+
 
 class WorkflowMacro(WikiMacroBase):
     _domain = 'messages'
@@ -438,7 +575,7 @@ class WorkflowMacro(WikiMacroBase):
     current ticket workflow is rendered. In WikiProcessors mode the `width`
     and `height` arguments can be specified.
 
-    (Defaults: `width = 800` and `heigth = 600`)
+    (Defaults: `width = 800` and `height = 600`)
 
     Examples:
     {{{
@@ -452,6 +589,14 @@ class WorkflowMacro(WikiMacroBase):
         leave.operations = leave_status
         leave.default = 1
 
+        create = <none> -> new
+        create.default = 1
+
+        create_and_assign = <none> -> assigned
+        create_and_assign.label = assign
+        create_and_assign.permissions = TICKET_MODIFY
+        create_and_assign.operations = may_set_owner
+
         accept = new,assigned,accepted,reopened -> accepted
         accept.permissions = TICKET_MODIFY
         accept.operations = set_owner_to_self
@@ -482,7 +627,7 @@ class WorkflowMacro(WikiMacroBase):
             parser = RawConfigParser()
             try:
                 parser.readfp(StringIO(text))
-            except ParsingError, e:
+            except ParsingError as e:
                 return system_message(_("Error parsing workflow."),
                                       unicode(e))
             raw_actions = list(parser.items('ticket-workflow'))
@@ -491,8 +636,7 @@ class WorkflowMacro(WikiMacroBase):
             [state for action in actions.itervalues()
                    for state in action['oldstates']] +
             [action['newstate'] for action in actions.itervalues()]))
-        action_labels = [attrs.get('name') or name
-                         for name, attrs in actions.items()]
+        action_labels = [attrs['label'] for attrs in actions.values()]
         action_names = actions.keys()
         edges = []
         for name, action in actions.items():
