@@ -14,16 +14,17 @@
 #
 # Author: Christopher Lenz <cmlenz@gmx.de>
 
-from __future__ import with_statement
-
 import os.path
+from abc import ABCMeta, abstractmethod
+from datetime import datetime
 
 from trac.admin import AdminCommandError, IAdminCommandProvider, get_dir_list
-from trac.config import ConfigSection, ListOption, Option
+from trac.config import ConfigSection, Option
 from trac.core import *
 from trac.resource import IResourceManager, Resource, ResourceNotFound
+from trac.util import as_bool
 from trac.util.concurrency import threading
-from trac.util.datefmt import time_now
+from trac.util.datefmt import time_now, utc
 from trac.util.text import printout, to_unicode, exception_to_unicode
 from trac.util.translation import _
 from trac.web.api import IRequestFilter
@@ -39,6 +40,10 @@ class InvalidRepository(TracError):
     """Exception raised when a repository is invalid."""
 
 
+class InvalidConnector(TracError):
+    """Exception raised when a repository connector is invalid."""
+
+
 class IRepositoryConnector(Interface):
     """Provide support for a specific version control system."""
 
@@ -117,7 +122,7 @@ class DbRepositoryProvider(Component):
     implements(IRepositoryProvider, IAdminCommandProvider)
 
     repository_attrs = ('alias', 'description', 'dir', 'hidden', 'name',
-                        'type', 'url')
+                        'sync_per_request', 'type', 'url')
 
     # IRepositoryProvider methods
 
@@ -134,6 +139,7 @@ class DbRepositoryProvider(Component):
             if 'name' in info and ('dir' in info or 'alias' in info):
                 info['id'] = id
                 reponames[info['name']] = info
+            info['sync_per_request'] = as_bool(info.get('sync_per_request'))
         return reponames.iteritems()
 
     # IAdminCommandProvider methods
@@ -291,6 +297,8 @@ class DbRepositoryProvider(Component):
                     continue
                 if k in ('alias', 'name') and is_default(v):
                     v = ''
+                if k in ('hidden', 'sync_per_request'):
+                    v = '1' if as_bool(v) else None
                 if k == 'dir' and not os.path.isabs(v):
                     raise TracError(_("The repository directory must be "
                                       "absolute"))
@@ -311,6 +319,10 @@ class RepositoryManager(Component):
     implements(IRequestFilter, IResourceManager, IRepositoryProvider,
                ITemplateProvider)
 
+    changeset_realm = 'changeset'
+    source_realm = 'source'
+    repository_realm = 'repository'
+
     connectors = ExtensionPoint(IRepositoryConnector)
     providers = ExtensionPoint(IRepositoryProvider)
     change_listeners = ExtensionPoint(IRepositoryChangeListener)
@@ -329,35 +341,15 @@ class RepositoryManager(Component):
 
         (''since 0.12'')""")
 
-    repository_type = Option('trac', 'repository_type', 'svn',
-        """Default repository connector type. (''since 0.10'')
+    default_repository_type = Option('versioncontrol',
+                                     'default_repository_type', 'svn',
+        """Default repository connector type.
 
-        This is also used as the default repository type for repositories
-        defined in [[TracIni#repositories-section repositories]] or using the
-        "Repositories" admin panel. (''since 0.12'')
+        This is used as the default repository type for repositories defined
+        in the [[TracIni#repositories-section repositories]] section or using
+        the "Repositories" admin panel. (''since 0.12'')
         """)
 
-    repository_dir = Option('trac', 'repository_dir', '',
-        """Path to the default repository. This can also be a relative path
-        (''since 0.11'').
-
-        This option is deprecated, and repositories should be defined in the
-        [TracIni#repositories-section repositories] section, or using the
-        "Repositories" admin panel. (''since 0.12'')""")
-
-    repository_sync_per_request = ListOption('trac',
-        'repository_sync_per_request', '(default)',
-        doc="""List of repositories that should be synchronized on every page
-        request.
-
-        Leave this option empty if you have set up post-commit hooks calling
-        `trac-admin $ENV changeset added` on all your repositories
-        (recommended). Otherwise, set it to a comma-separated list of
-        repository names. Note that this will negatively affect performance,
-        and will prevent changeset listeners from receiving events from the
-        repositories specified here. The default is to synchronize the default
-        repository, for backward compatibility. (''since 0.12'')""")
-
     def __init__(self):
         self._cache = {}
         self._lock = threading.Lock()
@@ -369,42 +361,38 @@ class RepositoryManager(Component):
     def pre_process_request(self, req, handler):
         from trac.web.chrome import Chrome, add_warning
         if handler is not Chrome(self.env):
-            for reponame in self.repository_sync_per_request:
+            for repo_info in self.get_all_repositories().values():
+                if not as_bool(repo_info.get('sync_per_request')):
+                    continue
                 start = time_now()
-                if is_default(reponame):
-                    reponame = ''
+                repo_name = repo_info['name'] or '(default)'
                 try:
-                    repo = self.get_repository(reponame)
-                    if repo:
-                        repo.sync()
-                    else:
-                        self.log.warning("Unable to find repository '%s' for "
-                                         "synchronization",
-                                         reponame or '(default)')
-                        continue
-                except TracError, e:
+                    repo = self.get_repository(repo_info['name'])
+                    repo.sync()
+                except InvalidConnector:
+                    continue
+                except TracError as e:
                     add_warning(req,
                         _("Can't synchronize with repository \"%(name)s\" "
                           "(%(error)s). Look in the Trac log for more "
-                          "information.", name=reponame or '(default)',
+                          "information.", name=repo_name,
                           error=to_unicode(e)))
-                except Exception, e:
+                except Exception as e:
                     add_warning(req,
                         _("Failed to sync with repository \"%(name)s\": "
                           "%(error)s; repository information may be out of "
                           "date. Look in the Trac log for more information "
                           "including mitigation strategies.",
-                          name=reponame or '(default)', error=to_unicode(e)))
+                          name=repo_name, error=to_unicode(e)))
                     self.log.error(
                         "Failed to sync with repository \"%s\"; You may be "
                         "able to reduce the impact of this issue by "
-                        "configuring [trac] repository_sync_per_request; see "
+                        "configuring the sync_per_request option; see "
                         "http://trac.edgewall.org/wiki/TracRepositoryAdmin"
-                        "#ExplicitSync for more detail: %s",
-                        reponame or '(default)',
+                        "#ExplicitSync for more detail: %s", repo_name,
                         exception_to_unicode(e, traceback=True))
                 self.log.info("Synchronized '%s' repository in %0.2f seconds",
-                              reponame or '(default)', time_now() - start)
+                              repo_name, time_now() - start)
         return handler
 
     def post_process_request(self, req, template, data, content_type):
@@ -413,12 +401,12 @@ class RepositoryManager(Component):
     # IResourceManager methods
 
     def get_resource_realms(self):
-        yield 'changeset'
-        yield 'source'
-        yield 'repository'
+        yield self.changeset_realm
+        yield self.source_realm
+        yield self.repository_realm
 
     def get_resource_description(self, resource, format=None, **kwargs):
-        if resource.realm == 'changeset':
+        if resource.realm == self.changeset_realm:
             parent = resource.parent
             reponame = parent and parent.id
             id = resource.id
@@ -426,7 +414,7 @@ class RepositoryManager(Component):
                 return _("Changeset %(rev)s in %(repo)s", rev=id, repo=reponame)
             else:
                 return _("Changeset %(rev)s", rev=id)
-        elif resource.realm == 'source':
+        elif resource.realm == self.source_realm:
             parent = resource.parent
             reponame = parent and parent.id
             id = resource.id
@@ -448,43 +436,43 @@ class RepositoryManager(Component):
             # TRANSLATOR: file /path/to/file.py at version 13 in reponame
             return _('%(kind)s %(id)s%(at_version)s%(in_repo)s',
                      kind=kind, id=id, at_version=version, in_repo=in_repo)
-        elif resource.realm == 'repository':
+        elif resource.realm == self.repository_realm:
             if not resource.id:
                 return _("Default repository")
             return _("Repository %(repo)s", repo=resource.id)
 
     def get_resource_url(self, resource, href, **kwargs):
-        if resource.realm == 'changeset':
+        if resource.realm == self.changeset_realm:
             parent = resource.parent
             return href.changeset(resource.id, parent and parent.id or None)
-        elif resource.realm == 'source':
+        elif resource.realm == self.source_realm:
             parent = resource.parent
             return href.browser(parent and parent.id or None, resource.id,
                                 rev=resource.version or None)
-        elif resource.realm == 'repository':
+        elif resource.realm == self.repository_realm:
             return href.browser(resource.id or None)
 
     def resource_exists(self, resource):
-        if resource.realm == 'repository':
+        if resource.realm == self.repository_realm:
             reponame = resource.id
         else:
             reponame = resource.parent.id
         repos = self.env.get_repository(reponame)
         if not repos:
             return False
-        if resource.realm == 'changeset':
+        if resource.realm == self.changeset_realm:
             try:
                 repos.get_changeset(resource.id)
                 return True
             except NoSuchChangeset:
                 return False
-        elif resource.realm == 'source':
+        elif resource.realm == self.source_realm:
             try:
                 repos.get_node(resource.id, resource.version)
                 return True
             except NoSuchNode:
                 return False
-        elif resource.realm == 'repository':
+        elif resource.realm == self.repository_realm:
             return True
 
     # IRepositoryProvider methods
@@ -497,13 +485,10 @@ class RepositoryManager(Component):
         """
         repositories = self.repositories_section
         reponames = {}
-        # eventually add pre-0.12 default repository
-        if self.repository_dir:
-            reponames[''] = {'dir': self.repository_dir}
         # first pass to gather the <name>.dir entries
         for option in repositories:
             if option.endswith('.dir') and repositories.get(option):
-                reponames[option[:-4]] = {}
+                reponames[option[:-4]] = {'sync_per_request': False}
         # second pass to gather aliases
         for option in repositories:
             alias = repositories.get(option)
@@ -586,6 +571,8 @@ class RepositoryManager(Component):
         :return: if no corresponding repository was defined,
                  simply return `None`.
 
+        :raises InvalidConnector: if the repository connector cannot be
+                                  opened.
         :raises InvalidRepository: if the repository cannot be opened.
         """
         reponame = reponame or ''
@@ -596,7 +583,7 @@ class RepositoryManager(Component):
         rdir = repoinfo.get('dir')
         if not rdir:
             return None
-        rtype = repoinfo.get('type') or self.repository_type
+        rtype = repoinfo.get('type') or self.default_repository_type
 
         # get a Repository for the reponame (use a thread-level cache)
         with self.env.db_transaction: # prevent possible deadlock, see #4465
@@ -646,7 +633,8 @@ class RepositoryManager(Component):
         hierarchy and return the name of its associated repository.
         """
         while context:
-            if context.resource.realm in ('source', 'changeset') and \
+            if context.resource.realm in (self.source_realm,
+                                          self.changeset_realm) and \
                     context.resource.parent:
                 return context.resource.parent.id
             context = context.parent
@@ -722,17 +710,13 @@ class RepositoryManager(Component):
         errors = []
         for repos in sorted(repositories, key=lambda r: r.reponame):
             reponame = repos.reponame or '(default)'
-            if reponame in self.repository_sync_per_request:
-                self.log.warn("Repository '%s' should be removed from [trac] "
-                              "repository_sync_per_request for explicit "
-                              "synchronization", reponame)
             repos.sync()
             for rev in revs:
                 args = []
                 if event == 'changeset_modified':
                     try:
                         old_changeset = repos.sync_changeset(rev)
-                    except NoSuchChangeset, e:
+                    except NoSuchChangeset as e:
                         errors.append(exception_to_unicode(e))
                         self.log.warn(
                             "No changeset '%s' found in repository '%s'. "
@@ -747,7 +731,7 @@ class RepositoryManager(Component):
                     try:
                         repos.sync_changeset(rev)
                         changeset = repos.get_changeset(rev)
-                    except NoSuchChangeset, e:
+                    except NoSuchChangeset as e:
                         errors.append(exception_to_unicode(e))
                         self.log.warn(
                             "No changeset '%s' found in repository '%s'. "
@@ -792,12 +776,12 @@ class RepositoryManager(Component):
             if prio >= 0: # no error condition
                 return connector
             else:
-                raise TracError(
+                raise InvalidConnector(
                     _('Unsupported version control system "%(name)s"'
                       ': %(error)s', name=rtype,
                       error=to_unicode(connector.error)))
         else:
-            raise TracError(
+            raise InvalidConnector(
                 _('Unsupported version control system "%(name)s": '
                   'Can\'t find an appropriate component, maybe the '
                   'corresponding plugin was not enabled? ', name=rtype))
@@ -824,10 +808,18 @@ class NoSuchNode(ResourceNotFound):
 class Repository(object):
     """Base class for a repository provided by a version control system."""
 
+    __metaclass__ = ABCMeta
+
     has_linear_changesets = False
 
     scope = '/'
 
+    realm = RepositoryManager.repository_realm
+
+    @property
+    def resource(self):
+        return Resource(self.realm, self.reponame)
+
     def __init__(self, name, params, log):
         """Initialize a repository.
 
@@ -847,11 +839,15 @@ class Repository(object):
         self.reponame = params['name']
         self.id = params['id']
         self.log = log
-        self.resource = Resource('repository', self.reponame)
 
+    def __repr__(self):
+        return '<%s %r %r %r>' % (self.__class__.__name__,
+                                  self.id, self.name, self.scope)
+
+    @abstractmethod
     def close(self):
         """Close the connection to the repository."""
-        raise NotImplementedError
+        pass
 
     def get_base(self):
         """Return the name of the base repository for this repository.
@@ -908,9 +904,10 @@ class Repository(object):
         """
         return None
 
+    @abstractmethod
     def get_changeset(self, rev):
         """Retrieve a Changeset corresponding to the given revision `rev`."""
-        raise NotImplementedError
+        pass
 
     def get_changeset_uid(self, rev):
         """Return a globally unique identifier for the ''rev'' changeset.
@@ -942,6 +939,7 @@ class Repository(object):
         except TracError:
             return False
 
+    @abstractmethod
     def get_node(self, path, rev=None):
         """Retrieve a Node from the repository at the given path.
 
@@ -951,18 +949,21 @@ class Repository(object):
         revision is returned, otherwise the Node corresponding to the youngest
         revision is returned.
         """
-        raise NotImplementedError
+        pass
 
+    @abstractmethod
     def get_oldest_rev(self):
         """Return the oldest revision stored in the repository."""
-        raise NotImplementedError
+        pass
     oldest_rev = property(lambda self: self.get_oldest_rev())
 
+    @abstractmethod
     def get_youngest_rev(self):
         """Return the youngest revision in the repository."""
-        raise NotImplementedError
+        pass
     youngest_rev = property(lambda self: self.get_youngest_rev())
 
+    @abstractmethod
     def previous_rev(self, rev, path=''):
         """Return the revision immediately preceding the specified revision.
 
@@ -971,8 +972,9 @@ class Repository(object):
 
         In presence of multiple parents, this follows the first parent.
         """
-        raise NotImplementedError
+        pass
 
+    @abstractmethod
     def next_rev(self, rev, path=''):
         """Return the revision immediately following the specified revision.
 
@@ -981,33 +983,42 @@ class Repository(object):
 
         In presence of multiple children, this follows the first child.
         """
-        raise NotImplementedError
+        pass
 
     def parent_revs(self, rev):
         """Return a list of parents of the specified revision."""
         parent = self.previous_rev(rev)
         return [parent] if parent is not None else []
 
+    @abstractmethod
     def rev_older_than(self, rev1, rev2):
         """Provides a total order over revisions.
 
         Return `True` if `rev1` is an ancestor of `rev2`.
         """
-        raise NotImplementedError
+        pass
 
+    # @abstractmethod
     def get_path_history(self, path, rev=None, limit=None):
         """Retrieve all the revisions containing this path.
 
         If given, `rev` is used as a starting point (i.e. no revision
         ''newer'' than `rev` should be returned).
         The result format should be the same as the one of Node.get_history()
+
+        :since 1.1.2: The method should be implemented in subclasses since
+                      it will be made abstract in Trac 1.3.1. A `TypeError`
+                      will result when instantiating classes that don't
+                      implement the method.
         """
         raise NotImplementedError
 
+    @abstractmethod
     def normalize_path(self, path):
         """Return a canonical representation of path in the repos."""
-        raise NotImplementedError
+        pass
 
+    @abstractmethod
     def normalize_rev(self, rev):
         """Return a (unique) canonical representation of a revision.
 
@@ -1021,26 +1032,32 @@ class Repository(object):
 
         :raise NoSuchChangeset: If the given `rev` isn't found.
         """
-        raise NotImplementedError
+        pass
 
     def short_rev(self, rev):
-        """Return a compact representation of a revision in the repos.
+        """Return a compact string representation of a revision in the
+        repos.
 
         :raise NoSuchChangeset: If the given `rev` isn't found.
+        :since 1.2: Always returns a string or `None`.
         """
-        return self.normalize_rev(rev)
+        norm_rev = self.normalize_rev(rev)
+        return str(norm_rev) if norm_rev is not None else norm_rev
 
     def display_rev(self, rev):
-        """Return a representation of a revision in the repos for displaying to
-        the user.
+        """Return a string representation of a revision in the repos for
+        displaying to the user.
 
-        This can be a shortened revision string, e.g. for repositories using
-        long hashes.
+        This can be a shortened revision string, e.g. for repositories
+        using long hashes.
 
         :raise NoSuchChangeset: If the given `rev` isn't found.
+        :since 1.2: Always returns a string or `None`.
         """
-        return self.normalize_rev(rev)
+        norm_rev = self.normalize_rev(rev)
+        return str(norm_rev) if norm_rev is not None else norm_rev
 
+    @abstractmethod
     def get_changes(self, old_path, old_rev, new_path, new_rev,
                     ignore_ancestry=1):
         """Generates changes corresponding to generalized diffs.
@@ -1051,7 +1068,7 @@ class Repository(object):
         The old_node is assumed to be None when the change is an ADD,
         the new_node is assumed to be None when the change is a DELETE.
         """
-        raise NotImplementedError
+        pass
 
     def is_viewable(self, perm):
         """Return True if view permission is granted on the repository."""
@@ -1063,12 +1080,16 @@ class Repository(object):
 class Node(object):
     """Represents a directory or file in the repository at a given revision."""
 
+    __metaclass__ = ABCMeta
+
     DIRECTORY = "dir"
     FILE = "file"
 
-    resource = property(lambda self: Resource('source', self.path,
-                                              version=self.rev,
-                                              parent=self.repos.resource))
+    realm = RepositoryManager.source_realm
+
+    @property
+    def resource(self):
+        return Resource(self.realm, self.path, self.rev, self.repos.resource)
 
     # created_path and created_rev properties refer to the Node "creation"
     # in the Subversion meaning of a Node in a versioned tree (see #3340).
@@ -1086,13 +1107,20 @@ class Node(object):
         self.rev = rev
         self.kind = kind
 
+    def __repr__(self):
+        name = u'%s:%s' % (self.repos.name, self.path)
+        if self.rev is not None:
+            name += '@' + unicode(self.rev)
+        return '<%s %r>' % (self.__class__.__name__, name)
+
+    @abstractmethod
     def get_content(self):
         """Return a stream for reading the content of the node.
 
         This method will return `None` for directories.
         The returned object must support a `read([len])` method.
         """
-        raise NotImplementedError
+        pass
 
     def get_processed_content(self, keyword_substitution=True, eol_hint=None):
         """Return a stream for reading the content of the node, with some
@@ -1111,14 +1139,16 @@ class Node(object):
         """
         return self.get_content()
 
+    @abstractmethod
     def get_entries(self):
         """Generator that yields the immediate child entries of a directory.
 
         The entries are returned in no particular order.
         If the node is a file, this method returns `None`.
         """
-        raise NotImplementedError
+        pass
 
+    @abstractmethod
     def get_history(self, limit=None):
         """Provide backward history for this Node.
 
@@ -1131,7 +1161,7 @@ class Node(object):
 
         :param limit: if given, yield at most ``limit`` results.
         """
-        raise NotImplementedError
+        pass
 
     def get_previous(self):
         """Return the change event corresponding to the previous revision.
@@ -1145,6 +1175,7 @@ class Node(object):
             else:
                 return p
 
+    @abstractmethod
     def get_annotations(self):
         """Provide detailed backward history for the content of this Node.
 
@@ -1152,37 +1183,41 @@ class Node(object):
         for that node.
         Only expected to work on (text) FILE nodes, of course.
         """
-        raise NotImplementedError
+        pass
 
+    @abstractmethod
     def get_properties(self):
         """Returns the properties (meta-data) of the node, as a dictionary.
 
         The set of properties depends on the version control system.
         """
-        raise NotImplementedError
+        pass
 
+    @abstractmethod
     def get_content_length(self):
         """The length in bytes of the content.
 
         Will be `None` for a directory.
         """
-        raise NotImplementedError
+        pass
     content_length = property(lambda self: self.get_content_length())
 
+    @abstractmethod
     def get_content_type(self):
         """The MIME type corresponding to the content, if known.
 
         Will be `None` for a directory.
         """
-        raise NotImplementedError
+        pass
     content_type = property(lambda self: self.get_content_type())
 
     def get_name(self):
         return self.path.split('/')[-1]
     name = property(lambda self: self.get_name())
 
+    @abstractmethod
     def get_last_modified(self):
-        raise NotImplementedError
+        pass
     last_modified = property(lambda self: self.get_last_modified())
 
     isdir = property(lambda self: self.kind == Node.DIRECTORY)
@@ -1199,6 +1234,8 @@ class Node(object):
 class Changeset(object):
     """Represents a set of changes committed at once in a repository."""
 
+    __metaclass__ = ABCMeta
+
     ADD = 'add'
     COPY = 'copy'
     DELETE = 'delete'
@@ -1210,8 +1247,11 @@ class Changeset(object):
     OTHER_CHANGES = (ADD, DELETE)
     ALL_CHANGES = DIFF_CHANGES + OTHER_CHANGES
 
-    resource = property(lambda self: Resource('changeset', self.rev,
-                                              parent=self.repos.resource))
+    realm = RepositoryManager.changeset_realm
+
+    @property
+    def resource(self):
+        return Resource(self.realm, self.rev, parent=self.repos.resource)
 
     def __init__(self, repos, rev, message, author, date):
         self.repos = repos
@@ -1220,6 +1260,10 @@ class Changeset(object):
         self.author = author or ''
         self.date = date
 
+    def __repr__(self):
+        name = u'%s@%s' % (self.repos.name, self.rev)
+        return '<%s %r>' % (self.__class__.__name__, name)
+
     def get_properties(self):
         """Returns the properties (meta-data) of the node, as a dictionary.
 
@@ -1231,6 +1275,7 @@ class Changeset(object):
         """
         return []
 
+    @abstractmethod
     def get_changes(self):
         """Generator that produces a tuple for every change in the changeset.
 
@@ -1243,7 +1288,7 @@ class Changeset(object):
         The `base_path` and `base_rev` are the source path and rev for the
         action (`None` and `-1` in the case of an ADD change).
         """
-        raise NotImplementedError
+        pass
 
     def get_branches(self):
         """Yield branches to which this changeset belong.
@@ -1260,6 +1305,13 @@ class Changeset(object):
         """
         return []
 
+    def get_bookmarks(self):
+        """Yield bookmarks associated with this changeset.
+
+        .. versionadded :: 1.1.5
+        """
+        return []
+
     def is_viewable(self, perm):
         """Return True if view permission is granted on the changeset."""
         return 'CHANGESET_VIEW' in perm(self.resource)
@@ -1267,6 +1319,20 @@ class Changeset(object):
     can_view = is_viewable  # 0.12 compatibility
 
 
+class EmptyChangeset(Changeset):
+    """Changeset that contains no changes. This is typically used when the
+    changeset can't be retrieved."""
+
+    def __init__(self, repos, rev, message=None, author=None, date=None):
+        if date is None:
+            date = datetime(1970, 1, 1, tzinfo=utc)
+        super(EmptyChangeset, self).__init__(repos, rev, message, author,
+                                             date)
+
+    def get_changes(self):
+        return iter([])
+
+
 # Note: Since Trac 0.12, Exception PermissionDenied class is gone,
 # and class Authorizer is gone as well.
 #
