@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2010 Edgewall Software
+# Copyright (C) 2003-2014 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2005-2007 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
@@ -15,32 +15,31 @@
 #
 # Author: Jonas Borgström <jonas@edgewall.com>
 
+import re
 from datetime import datetime, timedelta
 from fnmatch import fnmatchcase
-import re
 
 from genshi.builder import tag
 
-from trac.config import ListOption, BoolOption, Option
+from trac.config import BoolOption, ListOption, Option
 from trac.core import *
 from trac.mimeview.api import IHTMLPreviewAnnotator, Mimeview, is_binary
 from trac.perm import IPermissionRequestor, PermissionError
 from trac.resource import Resource, ResourceNotFound
 from trac.util import as_bool, embedded_numbers
-from trac.util.datefmt import http_date, to_datetime, utc
-from trac.util.html import escape, Markup
+from trac.util.datefmt import datetime_now, http_date, to_datetime, utc
+from trac.util.html import Markup, escape
 from trac.util.text import exception_to_unicode, shorten_line
 from trac.util.translation import _, cleandoc_
+from trac.versioncontrol.api import NoSuchChangeset, RepositoryManager
+from trac.versioncontrol.web_ui.util import *
 from trac.web.api import IRequestHandler, RequestDone
 from trac.web.chrome import (INavigationContributor, add_ctxtnav, add_link,
                              add_script, add_stylesheet, prevnext_nav,
                              web_context)
-from trac.wiki.api import IWikiSyntaxProvider, IWikiMacroProvider, parse_args
+from trac.wiki.api import IWikiMacroProvider, IWikiSyntaxProvider, parse_args
 from trac.wiki.formatter import format_to_html, format_to_oneliner
 
-from ..api import NoSuchChangeset, RepositoryManager
-from trac.versioncontrol.web_ui.util import * # `from .util import *` FIXME 2.6
-
 
 CHUNK_SIZE = 4096
 
@@ -118,15 +117,13 @@ class WikiPropertyRenderer(Component):
                                  'trac:description',
         doc="""Comma-separated list of version control properties to render
         as wiki content in the repository browser.
-
-        (''since 0.11'')""")
+        """)
 
     oneliner_properties = ListOption('browser', 'oneliner_properties',
                                  'trac:summary',
         doc="""Comma-separated list of version control properties to render
         as oneliner wiki content in the repository browser.
-
-        (''since 0.11'')""")
+        """)
 
     def match_property(self, name, mode):
         return 4 if name in self.wiki_properties \
@@ -180,6 +177,8 @@ class BrowserModule(Component):
 
     property_renderers = ExtensionPoint(IPropertyRenderer)
 
+    realm = RepositoryManager.source_realm
+
     downloadable_paths = ListOption('browser', 'downloadable_paths',
                                     '/trunk, /branches/*, /tags/*',
         doc="""List of repository paths that can be downloaded.
@@ -191,14 +190,14 @@ class BrowserModule(Component):
         repository name if the path does not point to the default repository
         (e.g. /reponame/trunk). Note that a simple prefix matching is
         performed on the paths, so aliases won't get automatically resolved.
-        (''since 0.10'')""")
+        """)
 
     color_scale = BoolOption('browser', 'color_scale', True,
         doc="""Enable colorization of the ''age'' column.
 
         This uses the same color scale as the source code annotation:
         blue is older, red is newer.
-        (''since 0.11'')""")
+        """)
 
     NEWEST_COLOR = (255, 136, 136)
 
@@ -206,7 +205,7 @@ class BrowserModule(Component):
         doc="""(r,g,b) color triple to use for the color corresponding
         to the newest color, for the color scale used in ''blame'' or
         the browser ''age'' column if `color_scale` is enabled.
-        (''since 0.11'')""")
+        """)
 
     OLDEST_COLOR = (136, 136, 255)
 
@@ -214,13 +213,13 @@ class BrowserModule(Component):
         doc="""(r,g,b) color triple to use for the color corresponding
         to the oldest color, for the color scale used in ''blame'' or
         the browser ''age'' column if `color_scale` is enabled.
-        (''since 0.11'')""")
+        """)
 
     intermediate_point = Option('browser', 'intermediate_point', '',
         doc="""If set to a value between 0 and 1 (exclusive), this will be the
         point chosen to set the `intermediate_color` for interpolating
         the color value.
-        (''since 0.11'')""")
+        """)
 
     intermediate_color = Option('browser', 'intermediate_color', '',
         doc="""(r,g,b) color triple to use for the color corresponding
@@ -228,7 +227,7 @@ class BrowserModule(Component):
         for the color scale (see `intermediate_point`).
         If not set, the intermediate color between `oldest_color` and
         `newest_color` will be used.
-        (''since 0.11'')""")
+        """)
 
     render_unsafe_content = BoolOption('browser', 'render_unsafe_content',
                                         'false',
@@ -240,12 +239,12 @@ class BrowserModule(Component):
         scripting attacks.
 
         For open repositories where anyone can check-in a file, it is
-        recommended to leave this option disabled (which is the default).""")
+        recommended to leave this option disabled.""")
 
     hidden_properties = ListOption('browser', 'hide_properties', 'svk:merge',
         doc="""Comma-separated list of version control properties to hide from
         the repository browser.
-        (''since 0.9'')""")
+        """)
 
     # public methods
 
@@ -332,13 +331,12 @@ class BrowserModule(Component):
             req.redirect(presel)
 
         path = req.args.get('path', '/')
-        rev = req.args.get('rev', '')
+        rev = req.args.getfirst('rev', '')
         if rev.lower() in ('', 'head'):
             rev = None
         format = req.args.get('format')
         order = req.args.get('order', 'name').lower()
         desc = 'desc' in req.args
-        xhr = req.get_header('X-Requested-With') == 'XMLHttpRequest'
 
         rm = RepositoryManager(self.env)
         all_repositories = rm.get_all_repositories()
@@ -374,7 +372,7 @@ class BrowserModule(Component):
                 # as a special shortcut to the latest revision.
                 rev_or_latest = rev or repos.youngest_rev
                 node = get_existing_node(req, repos, path, rev_or_latest)
-            except NoSuchChangeset, e:
+            except NoSuchChangeset as e:
                 raise ResourceNotFound(e, _('Invalid changeset number'))
             if node:
                 try:
@@ -383,7 +381,7 @@ class BrowserModule(Component):
                 except NoSuchChangeset:
                     pass
 
-            context = context.child(repos.resource.child('source', path,
+            context = context.child(repos.resource.child(self.realm, path,
                                                    version=rev_or_latest))
             display_rev = repos.display_rev
 
@@ -417,7 +415,7 @@ class BrowserModule(Component):
                 raise ResourceNotFound(_("No node %(path)s", path=path))
 
         quickjump_data = properties_data = None
-        if node and not xhr:
+        if node and not req.is_xhr:
             properties_data = self.render_properties(
                     'browser', context, node.get_properties())
             quickjump_data = list(repos.get_quickjump_entries(rev))
@@ -436,9 +434,9 @@ class BrowserModule(Component):
             'quickjump_entries': quickjump_data,
             'wiki_format_messages': \
                 self.config['changeset'].getbool('wiki_format_messages'),
-            'xhr': xhr,
+            'xhr': req.is_xhr,  # Remove in 1.3.1
         }
-        if xhr: # render and return the content only
+        if req.is_xhr: # render and return the content only
             return 'dir_entries.html', data, None
 
         if dir_data or repo_data:
@@ -517,7 +515,7 @@ class BrowserModule(Component):
                 continue
             try:
                 repos = rm.get_repository(reponame)
-            except TracError, err:
+            except TracError as err:
                 entry = (reponame, repoinfo, None, None,
                          exception_to_unicode(err), None)
             else:
@@ -540,7 +538,7 @@ class BrowserModule(Component):
                 else:
                     entry = (reponame, repoinfo, None, None, u"\u2013", None)
             if entry[4] is not None:  # Check permission in case of error
-                root = Resource('repository', reponame).child('source', '/')
+                root = Resource('repository', reponame).child(self.realm, '/')
                 if 'BROWSER_VIEW' not in context.perm(root):
                     continue
             repositories.append(entry)
@@ -586,7 +584,7 @@ class BrowserModule(Component):
         if rev:
             newest = repos.get_changeset(rev).date
         else:
-            newest = datetime.now(req.tz)
+            newest = datetime_now(req.tz)
 
         # Color scale for the age column
         timerange = custom_colorizer = None
@@ -784,7 +782,7 @@ class BrowserModule(Component):
                     rendered = None
                 prop = {'name': name, 'value': value, 'rendered': rendered}
                 return prop
-            except Exception, e:
+            except Exception as e:
                 self.log.warning('Rendering failed for property %s with '
                                  'renderer %s: %s', name,
                                  renderer.__class__.__name__,
@@ -1020,7 +1018,7 @@ class BlameAnnotator(object):
             return
         rev = self.annotations[lineno-1]
         chgset = self.changesets[lineno-1]
-        path = self.paths.get(rev, None)
+        path = self.paths.get(rev)
         # Note: path will be None if copy/rename is not supported
         # by get_history
 
