@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2009 Edgewall Software
+# Copyright (C) 2004-2014 Edgewall Software
 # Copyright (C) 2004 Daniel Lundin <daniel@edgewall.com>
 # Copyright (C) 2004-2006 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2006 Jonas Borgström <jonas@edgewall.com>
@@ -18,19 +18,17 @@
 # Author: Daniel Lundin <daniel@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
-from __future__ import with_statement
+import re
 
-import sys
-import time
-
-from trac.admin.api import console_date_format, get_console_locale
-from trac.core import TracError, Component, implements
-from trac.util import hex_entropy
-from trac.util.text import print_table, to_utf8
-from trac.util.translation import _
+from trac.admin.api import AdminCommandError, IAdminCommandProvider, \
+                           console_date_format, get_console_locale
+from trac.core import Component, ExtensionPoint, TracError, implements
+from trac.util import as_bool, as_int, hex_entropy, lazy
 from trac.util.datefmt import get_datetime_format_hint, format_date, \
-                              parse_date, to_datetime, to_timestamp
-from trac.admin.api import IAdminCommandProvider, AdminCommandError
+                              parse_date, time_now, to_datetime, to_timestamp
+from trac.util.text import print_table
+from trac.util.translation import _
+from trac.web.api import IRequestHandler, is_valid_default_handler
 
 UPDATE_INTERVAL = 3600 * 24 # Update session last_visit time stamp after 1 day
 PURGE_AGE = 3600 * 24 * 90 # Purge session after 90 days idle
@@ -43,21 +41,50 @@ COOKIE_KEY = 'trac_session'
 #       accessing the tables in the same order within the transaction,
 #       first `session`, then `session_attribute`.
 
-class DetachedSession(dict):
-    def __init__(self, env, sid):
-        dict.__init__(self)
-        self.env = env
+class SessionDict(dict):
+
+    def __init__(self):
+        super(SessionDict, self).__init__()
+        self.authenticated = False
+        self.last_visit = 0
         self.sid = None
-        if sid:
-            self.get_session(sid, authenticated=True)
-        else:
-            self.authenticated = False
-            self.last_visit = 0
-            self._new = True
-            self._old = {}
 
     def __setitem__(self, key, value):
-        dict.__setitem__(self, key, unicode(value))
+        super(SessionDict, self).__setitem__(key, unicode(value))
+
+    def as_bool(self, key, default=None):
+        """Return the value as a boolean. Return `default` if
+        if an exception is raised while converting the value to a
+        boolean.
+
+        :param key: the name of the session attribute
+        :keyword default: the value to return if the parameter is not
+                          specified or an exception occurs converting
+                          the value to a boolean.
+
+        :since: 1.2
+        """
+        if key not in self:
+            return default
+        return as_bool(self[key], default)
+
+    def as_int(self, key, default=None, min=None, max=None):
+        """Return the value as an integer. Return `default` if
+        if an exception is raised while converting the value to an
+        integer.
+
+        :param key: the name of the session attribute
+        :keyword default: the value to return if the parameter is not
+                          specified or an exception occurs converting
+                          the value to an integer.
+        :keyword min: lower bound to which the value is limited
+        :keyword max: upper bound to which the value is limited
+
+        :since: 1.2
+        """
+        if key not in self:
+            return default
+        return as_int(self[key], default, min, max)
 
     def set(self, key, value, default=None):
         """Set a variable in the session, or remove it if it's equal to the
@@ -69,7 +96,18 @@ class DetachedSession(dict):
             if value == default:
                 self.pop(key, None)
                 return
-        dict.__setitem__(self, key, value)
+        super(SessionDict, self).__setitem__(key, value)
+
+
+class DetachedSession(SessionDict):
+
+    def __init__(self, env, sid):
+        super(DetachedSession, self).__init__()
+        self.env = env
+        self._new = True
+        self._old = {}
+        if sid:
+            self.get_session(sid, authenticated=True)
 
     def get_session(self, sid, authenticated=False):
         self.env.log.debug("Retrieving session for ID %r", sid)
@@ -104,7 +142,7 @@ class DetachedSession(dict):
             return
 
         authenticated = int(self.authenticated)
-        now = int(time.time())
+        now = int(time_now())
 
         # We can't do the session management in one big transaction,
         # as the intertwined changes to both the session and
@@ -221,17 +259,20 @@ class Session(DetachedSession):
         self.req.outcookie[COOKIE_KEY]['expires'] = expires
         if self.env.secure_cookies:
             self.req.outcookie[COOKIE_KEY]['secure'] = True
-        if sys.version_info >= (2, 6):
-            self.req.outcookie[COOKIE_KEY]['httponly'] = True
+        self.req.outcookie[COOKIE_KEY]['httponly'] = True
+
+    _valid_sid_re = re.compile(r'[_A-Za-z0-9]+\Z')
 
     def get_session(self, sid, authenticated=False):
         refresh_cookie = False
 
+        if not authenticated and not self._valid_sid_re.match(sid):
+            raise TracError(_("Session ID must be alphanumeric."))
         if self.sid and sid != self.sid:
             refresh_cookie = True
 
         super(Session, self).get_session(sid, authenticated)
-        if self.last_visit and time.time() - self.last_visit > UPDATE_INTERVAL:
+        if self.last_visit and time_now() - self.last_visit > UPDATE_INTERVAL:
             refresh_cookie = True
 
         # Refresh the session cookie if this is the first visit after a day
@@ -244,7 +285,7 @@ class Session(DetachedSession):
         assert new_sid, 'Session ID cannot be empty'
         if new_sid == self.sid:
             return
-        if not to_utf8(new_sid).isalnum():
+        if not self._valid_sid_re.match(new_sid):
             raise TracError(_("Session ID must be alphanumeric."),
                             _("Error renaming session"))
         with self.env.db_transaction as db:
@@ -285,7 +326,7 @@ class Session(DetachedSession):
                       """, (sid,))
             elif len(authenticated_flags) == 1:
                 if not authenticated_flags[0]:
-                    # Update the anomymous session records so the session ID
+                    # Update the anonymous session records so the session ID
                     # becomes the user name, and set the authenticated flag.
                     self.env.log.debug("Promoting anonymous session %s to "
                                        "authenticated session for user %s",
@@ -303,7 +344,7 @@ class Session(DetachedSession):
                 try:
                     db("""INSERT INTO session (sid, last_visit, authenticated)
                           VALUES (%s, %s, 1)
-                          """, (self.req.authname, int(time.time())))
+                          """, (self.req.authname, int(time_now())))
                 except self.env.db_exc.IntegrityError:
                     self.env.log.warning('Authenticated session for %s '
                                          'already exists', self.req.authname)
@@ -311,7 +352,7 @@ class Session(DetachedSession):
         self._new = False
 
         self.sid = sid
-        self.bake_cookie(0) # expire the cookie
+        self.bake_cookie(0)  # expire the cookie
 
 
 class SessionAdmin(Component):
@@ -319,10 +360,12 @@ class SessionAdmin(Component):
 
     implements(IAdminCommandProvider)
 
+    request_handlers = ExtensionPoint(IRequestHandler)
+
     def get_admin_commands(self):
         hints = {
-           'datetime': get_datetime_format_hint(get_console_locale(self.env)),
-           'iso8601': get_datetime_format_hint('iso8601'),
+            'datetime': get_datetime_format_hint(get_console_locale(self.env)),
+            'iso8601': get_datetime_format_hint('iso8601'),
         }
         yield ('session list', '[sid[:0|1]] [...]',
                """List the name and email for the given sids
@@ -346,7 +389,8 @@ class SessionAdmin(Component):
                default if no suffix is specified).""",
                None, self._do_add)
 
-        yield ('session set', '<name|email> <sid[:0|1]> <value>',
+        yield ('session set', '<name|email|default_handler> '
+                              '<sid[:0|1]> <value>',
                """Set the name or email attribute of the given sid
 
                An sid suffix ':0' operates on an unauthenticated session with
@@ -371,13 +415,19 @@ class SessionAdmin(Component):
                format.""" % hints,
                None, self._do_purge)
 
+    @lazy
+    def _valid_default_handlers(self):
+        return sorted(handler.__class__.__name__
+                      for handler in self.request_handlers
+                      if is_valid_default_handler(handler))
+
     def _split_sid(self, sid):
         if sid.endswith(':0'):
-            return (sid[:-2], 0)
+            return sid[:-2], 0
         elif sid.endswith(':1'):
-            return (sid[:-2], 1)
+            return sid[:-2], 1
         else:
-            return (sid, 1)
+            return sid, 1
 
     def _get_sids(self):
         rows = self.env.db_query("SELECT sid, authenticated FROM session")
@@ -390,7 +440,7 @@ class SessionAdmin(Component):
                    if sid not in ('anonymous', 'authenticated', '*'))
         rows = self.env.db_query("""
             SELECT DISTINCT s.sid, s.authenticated, s.last_visit,
-                            n.value, e.value
+                            n.value, e.value, h.value
             FROM session AS s
               LEFT JOIN session_attribute AS n
                 ON (n.sid=s.sid AND n.authenticated=s.authenticated
@@ -398,12 +448,18 @@ class SessionAdmin(Component):
               LEFT JOIN session_attribute AS e
                 ON (e.sid=s.sid AND e.authenticated=s.authenticated
                     AND e.name='email')
+              LEFT JOIN session_attribute AS h
+                ON (h.sid=s.sid AND h.authenticated=s.authenticated
+                    AND h.name='default_handler')
             ORDER BY s.sid, s.authenticated
             """)
-        for sid, authenticated, last_visit, name, email in rows:
+        for sid, authenticated, last_visit, name, email, handler in rows:
             if all_anon and not authenticated or all_auth and authenticated \
                     or (sid, authenticated) in sids:
-                yield (sid, authenticated, last_visit, name, email)
+                yield (sid, authenticated,
+                       format_date(to_datetime(last_visit),
+                                   console_date_format),
+                       name, email, handler)
 
     def _complete_list(self, args):
         all_sids = self._get_sids() + ['*', 'anonymous', 'authenticated']
@@ -422,19 +478,16 @@ class SessionAdmin(Component):
     def _do_list(self, *sids):
         if not sids:
             sids = ['*']
-        print_table([(r[0], r[1], format_date(to_datetime(r[2]),
-                                              console_date_format),
-                      r[3], r[4])
-                     for r in self._get_list(sids)],
-                    [_('SID'), _('Auth'), _('Last Visit'), _('Name'),
-                     _('Email')])
+        headers = (_("SID"), _("Auth"), _("Last Visit"), _("Name"),
+                   _("Email"), _("Default Handler"))
+        print_table(self._get_list(sids), headers)
 
     def _do_add(self, sid, name=None, email=None):
         sid, authenticated = self._split_sid(sid)
         with self.env.db_transaction as db:
             try:
                 db("INSERT INTO session VALUES (%s, %s, %s)",
-                   (sid, authenticated, int(time.time())))
+                   (sid, authenticated, int(time_now())))
             except Exception:
                 raise AdminCommandError(_("Session '%(sid)s' already exists",
                                           sid=sid))
@@ -447,9 +500,13 @@ class SessionAdmin(Component):
         self.env.invalidate_known_users_cache()
 
     def _do_set(self, attr, sid, val):
-        if attr not in ('name', 'email'):
+        if attr not in ('name', 'email', 'default_handler'):
             raise AdminCommandError(_("Invalid attribute '%(attr)s'",
                                       attr=attr))
+        if attr == 'default_handler':
+            if val and val not in self._valid_default_handlers:
+                raise AdminCommandError(_("Invalid default_handler '%(val)s'",
+                                          val=val))
         sid, authenticated = self._split_sid(sid)
         with self.env.db_transaction as db:
             if not db("""SELECT sid FROM session
